.myButton {
    background-color:#f9f9f9;
    -moz-border-radius:4px;
    -webkit-border-radius:4px;
    border-radius:4px;
    border:1px solid #aaa;
    display:inline-block;
    cursor:pointer;
    color:#666666;
    font-size:16px;
    text-align: center;
    text-decoration:none;
    width: 30px;
    max-width: 30px;
    height: 30px;
    max-height: 30px;
    margin: 2px 1px 2px 1px;
}
.myButton:disabled {
    background-color:#fff;
    color: #bbb;
    cursor:auto;
}
.myButton:hover:enabled {
    background-color:#e9e9e9;
}
.myButton:active:enabled {
    position:relative;
    top:1px;
}
.myButton:focus {
    outline:0;
}
.myButtonToggledOn, .myButtonToggledOn:hover:enabled{
    background-color:yellow;
    color:red;
}

.toolbar {
    display: inline-block;
    border: 2px solid #aaa;
    box-sizing: border-box;
}
.button-group {
    display: inline-block;
}
.button-group+.button-group {
    border-left: 2px solid #aaa;
    padding-left: 2px;
}

.dropdown {
    position: relative;
    display: inline-block;
}
.dropdown-content {
    display: none;
    position: absolute;
    background-color: #f9f9f9;
    min-width: 160px;
    box-shadow: 0px 8px 16px 0px rgba(0,0,0,0.2);
}
.dropdown-content span.fakeLink {
    font-family: helvetica, sans-serif;
    font-size: 16px;
    cursor: pointer;
    color: black;
    padding: 12px 16px;
    text-decoration: none;
    display: block;
}
.dropdown-content span.fakeLink:hover {
    background-color: #ddd
}
.dropdown:hover .dropdown-content {
    display: block;
    z-index: 100;
}
.dropdown:hover .dropbtn {
    background-color: #3e8e41;
}





body {
    font-family: helvetica, sans-serif;
}
.path_string {
    font-family: "Courier New", Courier;
    font-size: 18px;
    margin-left: 10px;
    margin-top: 20px;
}
.component_connectivity {
    font-family: helvetica, Courier;
    font-size: 18px;
    margin-left: 10px;
    margin-top: 20px;
}
.ptN2ChartClass {
    margin-left: 10px;
    margin-top: 20px;
}
.ptN2LegendClass {
    margin-left: 10px;
    margin-top: 60px;
}



/* The Modal (background) */
.modal {
    display: none; /* Hidden by default */
    position: fixed; /* Stay in place */
    z-index: 9999; /* Sit on top */
    padding-top: 100px; /* Location of the box */
    left: 0;
    top: 0;
    width: 100%; /* Full width */
    height: 100%; /* Full height */
    overflow: auto; /* Enable scroll if needed */
    background-color: rgb(0,0,0); /* Fallback color */
    background-color: rgba(0,0,0,0.4); /* Black w/ opacity */
}
/* Modal Content */
.modal-content {
    font-family: helvetica, sans-serif;
    font-size: 14px;
    position: relative;
    background-color: #fefefe;
    margin: auto;
    padding: 0;
    border: 1px solid #888;
    width: 80%;
    box-shadow: 0 4px 8px 0 rgba(0,0,0,0.2),0 6px 20px 0 rgba(0,0,0,0.19);
    -webkit-animation-name: animatetop;
    -webkit-animation-duration: 0.4s;
    animation-name: animatetop;
    animation-duration: 0.4s
}
/* Add Animation */
@-webkit-keyframes animatetop {
    from {top:-300px; opacity:0}
    to {top:0; opacity:1}
}
@keyframes animatetop {
    from {top:-300px; opacity:0}
    to {top:0; opacity:1}
}
/* The Close Button */
.close {
    color: white;
    float: right;
    font-size: 28px;
    font-weight: bold;
}
.close:hover,
.close:focus {
    color: #000;
    text-decoration: none;
    cursor: pointer;
}
.modal-header {
    font-family: helvetica, sans-serif;
    font-size: 20px;
    padding: 2px 16px;
    background-color: steelblue;
    color: white;
    margin: 2px;
    padding: 10px;
}
.modal-body {padding: 2px 16px;}
.modal-footer {
    font-family: helvetica, sans-serif;
    font-size: 12px;
    padding: 2px 16px;
    background-color: steelblue;
    color: white;
    margin: 2px;
    padding: 10px;
}

@font-face {
    /* Subset of fonts from Font Awesome v4.6.3 created using http://fontello.com/ */
    font-family: 'fontello';
    src: url('data:application/font-woff;charset=utf-8;base64,{{fontello}}') format('woff');
}            

[class^="icon-"]:before, [class*=" icon-"]:before {
    display: inline-block;
    font: normal normal normal 14px/1 fontello;
    font-size: inherit;
    text-rendering: auto;
    -webkit-font-smoothing: antialiased;
    -moz-osx-font-smoothing: grayscale;
}
/* makes the font 33%% larger relative to the icon container */
.bigger-font {
    font-size: 1.33333333em;
    line-height: 0.75em;
    vertical-align: -15%;
}

/* To get SVG and PNG images of these icon fonts, use https://icomoon.io/app/#/select */
.icon-home:before { content: '\e800'; }
.icon-left-big:before { content: '\e801'; }
.icon-right-big:before { content: '\e802'; }
.icon-up-big:before { content: '\e803'; }
.icon-resize-full:before { content: '\e804'; }
.icon-resize-small:before { content: '\e805'; }
.icon-text-height:before { content: '\e806'; }
.icon-resize-vertical:before { content: '\e807'; }
.icon-floppy:before { content: '\e808'; }
.icon-search:before { content: '\e809'; }
.icon-minus:before { content: '\e80a'; }
.icon-exchange:before { content: '\f0ec'; }
.icon-terminal:before { content: '\f120'; }
.icon-help:before { content: '\f128'; }
.icon-eraser:before { content: '\f12d'; }
.icon-sort-number-up:before { content: '\f162'; }
.icon-right:before { content: '\f178'; }
.icon-map-signs:before { content: '\f277'; }

.context-menu {
  display: none;
  position: absolute;
  z-index: 10;
  padding: 12px 0;
  width: 240px;
  background-color: #fff;
  border: solid 1px #dfdfdf;
  box-shadow: 1px 1px 2px #cfcfcf;
}

.context-menu--active {
  display: block;
}

.context-menu__items {
  list-style: none;
  margin: 0;
  padding: 0;
}

.context-menu__item {
  display: block;
  margin-bottom: 4px;
}

.context-menu__item:last-child {
  margin-bottom: 0;
}

.context-menu__link {
  display: block;
  padding: 4px 12px;
  color: #0066aa;
  text-decoration: none;
}

.context-menu__link:hover {
  color: #fff;
  background-color: #0066aa;
}

.tool-tip {
  z-index: 10;
  font-size: 11px;
  padding: 5;
  background-color: #fff;
  border: solid 1px #dfdfdf;   
}

div.offgrid {
  z-index: 10;
  font-size: 11px;
  padding: 5;
  background-color: #fff;
  border: solid 1px #dfdfdf;
  position: absolute;
  opacity: 0.7;
}
